
#include "gamemanager.h"
#include "constants.h"
#include "utils.h"

#include <algorithm>
#include <alleggl.h>

GameManager::GameManager() {
	trackObject = 0;
	tickPointer();
	
}

void GameManager::render()
{
	// Render background objects.
	for (GameObjectIterator i = things.begin(); i != things.end(); i++) {
		GameObject *obj = *i;
		obj->renderBackground();
	}
	glEnable(GL_DEPTH_TEST);
	// Render main scene
	for (GameObjectIterator i = things.begin(); i != things.end(); i++) {
		GameObject *obj = *i;
		obj->render();
	}
	glDisable(GL_DEPTH_TEST);
	// Render foreground overlay
	for (GameObjectIterator i = things.begin(); i != things.end(); i++) {
		GameObject *obj = *i;
		obj->renderForeground();
	}
	renderParticles();
	renderGoals();
	renderPointer();
}

static bool IsAlive(GameObject *obj)
{
	return ! obj->dead;
}

static bool IsDead(GameObject *obj)
{
	return obj->dead;
}

void GameManager::tickPointer()
{
	// get mouse_x and mouse_y
	pointerx = (float) mouse_x;
	pointery = (float) mouse_y;
	pointerx = pointerx / ScreenWidth;
	pointery = pointery / ScreenHeight;
	// centre them
	pointerx = pointerx - 0.5;
	// invert Y access to - our coords have y=0 at the bottom
	pointery = 0.5 - pointery; 
	// Scale back up to world coords.
	pointerx = pointerx * ViewWidth;
	pointery = pointery * ViewHeight;
	// Offset by centre
	pointerx = pointerx + viewCentre.x;
	pointery = pointery + viewCentre.y;
}

// This is called with perspective projection still active
void GameManager::renderPointer()
{
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glBindTexture(GL_TEXTURE_2D, 
		resourceManager.getTexture(ResourceManager::PointerTexture));
	glEnable(GL_TEXTURE_2D);
	
	const float pointersize = 0.2;
	float leftx = pointerx - pointersize;
	float bottomy = pointery - pointersize;
	float rightx = pointerx + pointersize;
	float topy = pointery + pointersize;
	glBegin(GL_QUADS);
	glColor3f(1,1,1);
	glTexCoord2f(0,0);
	glVertex3f(leftx,bottomy, 0);
	glTexCoord2f(1,0);
	glVertex3f(rightx,bottomy, 0);
	glTexCoord2f(1,1);
	glVertex3f(rightx,topy, 0);
	glTexCoord2f(0,1);
	glVertex3f(leftx,topy, 0);
	glEnd();
}

void GameManager::renderGoals()
{
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	for (GoalList::iterator i = goals.begin(); i != goals.end(); i++) {
		Goal & g = *i;
		float red = 0.5, green = 0.5, blue = 0.5;
		if (g.satisfied) {
			green = 1.0;
		} else {
			red = 1.0;
		}
		glColor3f(red,green,blue); // light green
		DrawCircle(g.location ,g.radius);
		showRadar(g.location, red,green,blue);
	}
}

void GameManager::resetGoals()
{
	// Init the goals stuff.
	for (GoalList::iterator i = goals.begin(); i != goals.end(); i++) {
		Goal & g = *i;
		g.satisfied = false;
	}
	numStars = 0;
	goalsComplete = 0;
}

void GameManager::tick()
{
	resetGoals();
	tickPointer();
	
	cleanList(deadThings);
	// Take any new objects and stick them into things.
	for (GameObjectIterator i = newThings.begin(); i != newThings.end(); i++) {
		things.push_back(*i);
	}
	newThings.clear(); // clear out newThings
	
	// Tick them.
	for (GameObjectIterator i = things.begin(); i != things.end(); i++) {
		GameObject *obj = *i;
		obj->tick();
	}
	collideTest();
	
	if (trackObject) {
		if (trackObject->dead) {
			trackObject = 0;
		} else {
			viewCentre = trackObject->location + viewOffset;
		}
	}
	// Clean out dead objects, move them into a dead list.
	std::remove_copy_if(things.begin(), things.end(), back_inserter(deadThings), IsAlive);
	things.erase(std::remove_if(things.begin(), things.end(), IsDead), things.end());
	// Tick the particles
	tickParticles();
	
}

void GameManager::collideTest()
{
	for (GameObjectIterator i = things.begin(); i != things.end(); i++) 
	{
		GameObject * obj1 = *i;
		if (! obj1->collidable) 
		{
			continue;
		}
		for (GameObjectIterator j = (i + 1); j != things.end(); j++) 
		{
			GameObject * obj2 = *j;
			if (! obj2->collidable) 
			{
				continue;
			}
			Vect3d relPos = obj2->location - obj1->location;
			double collideDist = obj1->radius + obj2->radius;
			if (relPos.magnitudeSquared() < (collideDist * collideDist)) 
			{
				handleCollision(obj1, obj2);
			}
		}
	}
}

void GameManager::handleCollision(GameObject *obj1, GameObject * obj2)
{
	obj1->collided(obj2);
	obj2->collided(obj1);
}

void GameManager::cleanList( GameObjectList &list)
{
	for (GameObjectIterator i = list.begin(); i != list.end(); i++) {
		delete(*i);
	}
	list.clear();
}

void GameManager::triggerSound(ResourceManager::SampleType sample, const Vect3d &where)
{
	SAMPLE *s = resourceManager.getSample(sample);
	play_sample(s, 192 /* vol */, 128 /* pan */, 1000, 0);
}

GameObject * GameManager::findNearestCollidableExcept(Vect3d where, GameObject * exceptWhat)
{
	Component minDistSquared = 1e9; // A big number!
	GameObject * nearest = 0;
	for (GameObjectIterator i = things.begin(); i != things.end(); i++) {
		GameObject *obj = *i;
		// Ignore the object we should ignore.
		if (obj == exceptWhat) {
			continue;
		}
		// Ignore non-collidable objects.
		if (! obj->collidable) {
			continue;
		}
		Component distSquared = (obj->location - where).magnitudeSquared();
		if ( distSquared < minDistSquared) {
			nearest = obj;
			minDistSquared = distSquared;
		}
	}
	return nearest;
}
		
void GameManager::registerStarPos(GameObject *star)
{
	numStars ++;
	// Find out if we're within any goals.
	for (GoalList::iterator i = goals.begin(); i != goals.end(); i++) {
		Goal & g = *i;
		if (g.satisfied) {
			continue;
		}
		// Work out distance.
		Vect3d relPos = g.location - star->location;
		Component targetDist = g.radius - star->radius;
		if (relPos.magnitudeSquared() < (targetDist * targetDist)) {
			g.satisfied = true;
			goalsComplete ++;
		}
	}
}

void GameManager::renderParticles()
{
	// Depth test should be off.
	glEnable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	for (ParticleList::iterator i = particles.begin(); i != particles.end(); i++) {
		Particle &p = *i;
		// Draw each particle as a triangle fan.
		float brightness = 1.0;
		if (p.life < (p.maxlife / 3)) {
			brightness = ((float) p.life) / ((float) p.maxlife) * 3.0;
		}
		glBegin(GL_TRIANGLE_FAN);
		glColor4f(p.r, p.g, p.b, brightness);
		glVertex3f(p.location.x, p.location.y, p.location.z);
		glColor4f(p.r, p.g, p.b, 0.0 /* Alpha */);
		glVertex3f(p.location.x, p.location.y + p.radius, p.location.z);
		glVertex3f(p.location.x - p.radius, p.location.y, p.location.z);
		glVertex3f(p.location.x, p.location.y - p.radius, p.location.z);
		glVertex3f(p.location.x + p.radius, p.location.y, p.location.z);
		glVertex3f(p.location.x, p.location.y + p.radius, p.location.z);
		glEnd();
	}
}

static bool IsParticleDead(const Particle &p)
{
	return (p.life <= 0);
}

void GameManager::tickParticles()
{
	// Create some test partciles.
	/*
	Particle testParticle(Vect3d(0,0,0));
	testParticle.velocity.x = 0.05 - FloatRand() * 0.1;
	testParticle.velocity.y = 0.05 - FloatRand() * 0.1;
	addParticle(testParticle);
	*/
	
	for (ParticleList::iterator i = particles.begin(); i != particles.end(); i++) {
		Particle &p = *i;
		p.location = p.location + p.velocity;
		p.life --;
	}
	// Remove dead ones.
	particles.erase(
		std::remove_if(particles.begin(), particles.end(), IsParticleDead), particles.end());
}

void GameManager::createExplosion(GameObject *obj)
{
	// Create a big, bright, stationary, shortlived particle.
	{
		Particle p(obj->location);
		p.radius = obj->radius * 2.0;
		p.life = 6;
		addParticle(p);
	}
	// Create some smaller, longer lasting particles.
	for (int i=0; i<10; i++) {
		Particle p(obj->location);
		p.velocity.x = 0.05 - FloatRand() * 0.1;
		p.velocity.y = 0.05 - FloatRand() * 0.1;
		addParticle(p);
	}
}

void GameManager::showRadar(Vect3d where, float r, float g, float b)
{
	// If the location is further than the screen away, show an 
	// arrow pointing to it.
	Vect3d rel = (where - viewCentre);
	const float MinRadarDist = (ViewWidth / 2.0);
	if (rel.magnitudeSquared() > (MinRadarDist * MinRadarDist)) {
		rel.normalise();
		Vect3d start = viewCentre + (rel * (ViewHeight * 0.5));
		Vect3d end = viewCentre + (rel * (ViewHeight * 0.45));
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		glDisable(GL_BLEND);
		glColor3f(r,g,b);
		glBegin(GL_LINES);
		glVertex3f(start.x,start.y,start.z);
		glVertex3f(end.x,end.y,end.z);
		glEnd();
	}
}




