/*
 * Main file
 *
 * (c) 2005 Mark Robson 
 * mark@vectrex.org.uk
 * Allegro XmasHack 2005 entry
 * 
 */
#include "model.h"
#include "resourcemanager.h"
#include "gamemanager.h"
#include "playerinfo.h"
#include "constants.h"
#include "utils.h"
#include "starfield.h"
#include "ship.h"
#include "star.h"
#include "leveldata.h"

#include <alleggl.h>
#include <GL/glu.h>
#include <stdexcept>
#include <cmath>
#include <cstring>

static bool fullscreen = false;

static int GoalTime = 300; // 6 sec

PlayerInfo playerInfo;

/*
 * Rendering context states that we will never change.
 */
static void InitGl()
{
	/* remove back faces */
	glEnable(GL_CULL_FACE);
  
	/* speedups */
	glDisable(GL_DITHER);
	glShadeModel(GL_FLAT);
	glPolygonMode(GL_FRONT , GL_FILL); // GL_FILL or GL_LINE.
}

static BITMAP *hudBmp;

static void InitHud()
{
	hudBmp = create_bitmap(512, 16);
}

static void RenderHud(GameManager & manager)
{
	int white = makecol(255,255,255);
	int black = 0;
	clear_to_color(hudBmp,black);
	textprintf_ex(hudBmp, resourceManager.getFont(), 0,0, white, black, 
		"COMPLETE: %d / %d  FUEL:%04d",
		manager.goalsComplete, manager.countGoals(),
		playerInfo.fuel);
	// Copy it to a texture, then render a quad.
	GLuint hudtex = allegro_gl_make_texture_ex(
		AGL_TEXTURE_FLIP,
		hudBmp,
		GL_LUMINANCE8);
	glBindTexture(GL_TEXTURE_2D, hudtex);
	// draw a quad.
	glColor3f(1,1,1);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2f(0,0);
	glTexCoord2f(1,0);
	glVertex2f(512,0);
	glTexCoord2f(1,1);
	glVertex2f(512,32);
	glTexCoord2f(0,1);
	glVertex2f(0,32);
	glEnd();
	// delete the texture.
	glDeleteTextures(1, &hudtex);	
}


/*
 * Called at beginning of frame - init projection matrix.
 */
static void InitProjection(GameManager &man)
{
	glMatrixMode(GL_MODELVIEW);
  	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
  	glLoadIdentity();
	// Set the frustum...
	// Calculate Y field of view.
	float fovy = 180 * (atan((ViewHeight / 2) / ViewDistance) / M_PI);
	fovy *= 2;
	gluPerspective(fovy, (GLfloat)ScreenWidth/(GLfloat)ScreenHeight, 
		ViewDistance / 2,
		ViewDistance * 3);
	glTranslatef(- man.viewCentre.x, - man.viewCentre.y, - ViewDistance - man.viewCentre.z);
		
	glMatrixMode(GL_MODELVIEW);
 	
	// glEnable(GL_LIGHTING);
	static GLfloat LightPosition[] = {1.0f, 1.5f, 1.0f, 0.0f}; // last number 0 = Infinity distance in that dir

	static GLfloat LightAmbient[]  = {0.4f, 0.4f, 0.4f, 1.0f};
	static GLfloat LightDiffuse[]  = {0.6f, 0.6f, 0.6f, 1.0f};
	glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);
	glLightfv(GL_LIGHT1, GL_POSITION, LightPosition);
	glEnable(GL_LIGHT1);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_NORMALIZE); // Normalise our normals please.
	
	glDepthFunc(GL_LESS);
	glClearDepth(ViewDistance * 3);
}

static void InitRender(GameManager &man)
{
	// Render background, clear buf etc..
	glClearColor(0,0,0.6, 0);
	glClear(GL_DEPTH_BUFFER_BIT);
	// Render background.
	SetOrthoProjection();
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glBindTexture(GL_TEXTURE_2D, 
		resourceManager.getTexture(ResourceManager::BackgroundTexture));
	glEnable(GL_TEXTURE_2D);
	// Make it look as if this texture is a long way off,
	// Even though it actually is at z=0
	double textureOffsetX = (man.viewCentre.x) / 20;
	double textureOffsetY = (man.viewCentre.y) / 20;
	double textureWidth = 1.0 * ((float)ScreenWidth / (float)ScreenHeight);
	glColor3f(0.4, 0.4, 0.4); // dark
	glBegin(GL_QUADS);
	glTexCoord2f(0 + textureOffsetX,0 + textureOffsetY);
	glVertex2f(0,0);
	glTexCoord2f(textureWidth + textureOffsetX,0 + textureOffsetY);
	glVertex2f(ScreenWidth,0);
	glTexCoord2f(textureWidth + textureOffsetX,1 + textureOffsetY);
	glVertex2f(ScreenWidth,ScreenHeight);
	glTexCoord2f(0 + textureOffsetX,1 + textureOffsetY);
	glVertex2f(0,ScreenHeight);
	glEnd();
	
	
}

static void InitAllegro()
{
	allegro_init();
	install_allegro_gl();

	allegro_gl_clear_settings();
	/* allegro_gl_set (AGL_Z_DEPTH, 16); */
	allegro_gl_set (AGL_COLOR_DEPTH, 16);
	allegro_gl_set (AGL_DOUBLEBUFFER, 1);
	allegro_gl_set (AGL_Z_DEPTH, 16);
	// TODO: Make fullscreen an option.
	if (fullscreen) {
		allegro_gl_set (AGL_FULLSCREEN, TRUE);	
	} else {
		allegro_gl_set (AGL_WINDOWED, TRUE);	
	}
	allegro_gl_set (AGL_SUGGEST, AGL_COLOR_DEPTH | AGL_Z_DEPTH
	              | AGL_DOUBLEBUFFER | AGL_WINDOWED);

	set_color_depth (16);
	if (set_gfx_mode(GFX_OPENGL, ScreenWidth, ScreenHeight, 0, 0) < 0) {
		throw std::runtime_error("Unable to set gfx mode");
	}
	int depthBits;
	glGetIntegerv(GL_DEPTH_BITS, & depthBits);
	if (depthBits == 0) {
		throw std::runtime_error("No depth buffer");
	}
	install_keyboard();
	install_mouse();
	install_timer();
	install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, "");
	InitGl();
}

// Init things which only need to be done once.
static void InitAll()
{
	InitAllegro();
	InitHud();
	
	resourceManager.loadAll();
}

static void Render(GameManager &man, bool showHud = true)
{
	InitRender(man);
	InitProjection(man);
	// DO main rendeirng here.
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	man.render();
	// ModelTest();
	// Set ortho for text.
	SetOrthoProjection();
	// Render HUD etc here.
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
	if (showHud) {
		RenderHud(man);
	}
	allegro_gl_flip();
}

static void CheckWinLose(GameManager &man)
{
	// determine if player is dead.
	// If we're dead, or ran out of fuel, we need to 
	if (playerInfo.dead) {
		playerInfo.failedTime ++;
	} else {
		if (playerInfo.fuel <= 0) {
			playerInfo.failedTime ++;
		}
	}
	if (man.goalsComplete == man.countGoals()) {
		playerInfo.completeTime ++;
	} else {
		playerInfo.completeTime = 0;
	}
	if (playerInfo.failedTime > GoalTime) {
		playerInfo.failed = true;
	}
	if (playerInfo.completeTime > GoalTime) {
		playerInfo.levelComplete = true;
	}
}

static bool LevelLoop(GameManager &man)
{
	bool oldPaused = false;
	bool paused = false;
	while (! (playerInfo.levelComplete || playerInfo.failed)) {
		if (key[KEY_LEFT]) {
			man.viewCentre.x -= 0.1;
		}
		if (key[KEY_RIGHT]) {
			man.viewCentre.x += 0.1;
		}
		if (key[KEY_UP]) {
			man.viewCentre.y += 0.1;
		}
		if (key[KEY_DOWN]) {
			man.viewCentre.y -= 0.1;
		}
		bool newPaused = key[KEY_PAUSE];
		if (newPaused && ! oldPaused ) {
			paused = ! paused;
		}
		oldPaused = newPaused;
		if (! paused) {
			man.tick();
			CheckWinLose(man);
		}
		Render(man);
		rest(20);
		if (key[KEY_ESC]) {
			return true;
		}
		if (key[KEY_F12]) {
			// cheat;
			playerInfo.levelComplete = true;
			return false;
		}
	}
	return false;
}

static void InitLevel(GameManager & man)
{
	man.cleanUp();
	
	StarField * starfield = new StarField();
	starfield->init();
	man.add(starfield);
	// Add the player's ship.
	Ship * ship = new Ship();
	man.add(ship);
	man.trackObject = ship;

	const LevelData & level = Levels[playerInfo.level - 1];
	// place stars
	for (int i=0; i< level.numStars; i++) {
		Star * star = new Star();
		star->location.x = level.stars[i].x;
		star->location.y = level.stars[i].y;
		man.add(star);
	}
	// Set up goals
	for (int i=0; i< level.numGoals; i++) {
		Vect3d where(level.goals[i].x, level.goals[i].y, 0.0);
		man.addGoal(where);
	}
	// Set fuel level
	playerInfo.maxfuel = level.maxfuel;
	playerInfo.restartLevel();
}

static void ShowWon()
{
	ShowIntermission("YOU WIN!");
}

void GameLoop()
{
	playerInfo = PlayerInfo();
	bool finished = false;
	while (! finished) {
		if (playerInfo.levelComplete) {
			playerInfo.level ++;
		}
		if (playerInfo.level > NumLevels) {
			// won!
			ShowWon();
			return;
		}
		char buf[200];
		std::sprintf(buf, "ENTERING LEVEL %d", playerInfo.level);
		ShowIntermission(buf);
		
		GameManager man;
		InitLevel(man);
		if (LevelLoop(man)) {
			finished = true;
		}
	}
}

static void MainLoop()
{
	// Show title screen
	// while not finished...
	GameLoop();
}

int main(int argc, char *argv[])
{
	try {
		if (argc>1) {
			if (argv[1][0] == 'f') {
				fullscreen = true;
			}
		}
		InitAll();
		MainLoop();
	} catch (std::runtime_error &e) 
	{
		allegro_message ("Unexpected exception:%s\nAllegro error: %s\n"
		                 "Allegro GL error : %s\n",
		                 e.what(), allegro_error, allegro_gl_error);
	} catch (...) {
		allegro_message ("Unknown exception thrown");
	}
	set_gfx_mode(GFX_TEXT, 0,0, 0, 0);
	return 0;
}

END_OF_MAIN()


