#include "ship.h"

#include "utils.h"
#include "playerinfo.h"

#include <alleggl.h>
#include <GL/glu.h>
#include <stdexcept>
#include <cmath>
#include <algorithm>

static const float ShipScale = (1.0 / 15.0);

static const float MinPointerDist = 0.5;
static const float AngleTolerance = 2.9; // Degrees
static const float TurnSpeed = 3.0; // Degrees per tick.

static const float EngineAccel = 0.004; // units per tick per tick
static const float TractorAccel = 0.001; // Amount it affects the ship
static const float StarMassRatio = 4.0; // accel. of the star.
static const float TractorRange = 3.50;
static const float MinFrictionSpeed = 0.0001; // Speed below which an object stops dead.
static const float TractorBeamRadius = 0.6;
static const float TractorTextureSpeed = 0.01;

Ship::Ship()
{
	azimuth = 45.0;
	radius = 0.6;
	tractorOffset = 0;
}

void Ship::renderBackground()
{
	if (tractorOn) 
	{
		// Draw the tractor beam.
		glDisable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, 
			resourceManager.getTexture(ResourceManager::TractorTexture));
		glColor4f(1.0, 1.0, 1.0, 0.5);
		// Work out a triangle.
		Vect3d rel = tractorTarget - location;
		// Normalise and then rotate through 90 degrees.
		rel.normalise();
		std::swap(rel.x, rel.y);
		rel.y = - rel.y;
		rel = rel * TractorBeamRadius;
		
		glBegin(GL_TRIANGLES);
		glTexCoord2f(0.5, 0 + tractorOffset);
		glVertex3f(location.x,location.y,location.z);
		glTexCoord2f(1.0, 1.0 + tractorOffset);
		glVertex3f(tractorTarget.x + rel.x, tractorTarget.y + rel.y, tractorTarget.z);
		glTexCoord2f(0.0, 1.0 + tractorOffset);
		glVertex3f(tractorTarget.x - rel.x, tractorTarget.y - rel.y, tractorTarget.z);
		glEnd();
	}
}

void Ship::render()
{
	const Model &model = resourceManager.getModel(ResourceManager::Ship);
	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glColor3f(1.0, 1.0,1.0);
	glBindTexture(GL_TEXTURE_2D, 
		resourceManager.getTexture(ResourceManager::ShipTexture));
	glPushMatrix();
	glTranslatef(location.x, location.y, location.z);
	// Rotate ship about X axis 90 degrees so it's now pointing North (+y)
	// Rather than away (-z)
	// glRotatef(90.0, 1,0,0);
	// rotate ship to desired angle.
	// Note we're now rotating about the Z axis
	glRotatef(azimuth, 0,0,-1);
	glScalef( ShipScale, ShipScale, ShipScale);
	model.draw(true);
	glPopMatrix();
	// Draw a sphere to calibrate the ship radius.
	//glDisable(GL_TEXTURE_2D);
	//glPolygonMode(GL_FRONT , GL_LINE); // GL_FILL or GL_LINE.
	//DrawSphere(location , radius , 8 , 0 /* 0=quads, 1=tri */ ,
	//0 /*theta1 */, 2 * M_PI /*theta2 */,
	//	- (M_PI / 2) /* phi1 */ , (M_PI / 2) /* phi2 */);
	//glPolygonMode(GL_FRONT , GL_FILL); // GL_FILL or GL_LINE.
}

void Ship::applyTractorBeam()
{
	// Target the star *behind* the ship.
	Vect3d targetLocation = location - Vect3d::fromPolar(azimuth, radius);
	
	GameObject * target = manager->findNearestCollidableExcept(targetLocation, this);
	if (target == 0) {
		tractorOn = false;
		return;
	}
	// Get the normalised dist.
	Vect3d toTarget = target->location - location;
	if (toTarget.magnitudeSquared() < (radius * radius)) {
		tractorOn = false;
		// Too near!
		return;
	}
	// check max tractor range distance.
	if (toTarget.magnitudeSquared() > (TractorRange * TractorRange)) {
		tractorOn = false;
		// Too far
		return;
	}
	
	tractorTarget = target->location;
	
	toTarget.normalise();
	// Accelerate the target towards the ship, and vice versa.
	// Seeing as the star will be much heavier, accelerate it less.
	velocity = velocity + (toTarget * TractorAccel);
	
	target->velocity = target->velocity - (toTarget * (TractorAccel / StarMassRatio));
}

static const double ExhaustSpeed = 0.10;

void Ship::createExhaustParticles()
{
	Vect3d forwards = Vect3d::fromPolar(azimuth, 1.0);
	// Work out our forwards and right vectors.
	Vect3d right(forwards);
	std::swap(right.x, right.y);
	right.y = - right.y;
	
	Vect3d backwards = forwards * (-0.7 * radius);
	right = right * (radius * 0.2);
	
	Particle p(location + backwards + right);
	p.r = p.g = p.b = 0.5;
	p.radius = 0.10 + FloatRand() * 0.10;
	p.life = 15;
	p.velocity = velocity + (backwards * ExhaustSpeed);
	manager->addParticle(p);
	p.radius = 0.10 + FloatRand() * 0.10;
	p.location = location + backwards - right;
	p.velocity = velocity + (backwards * ExhaustSpeed);
	manager->addParticle(p);
}

void Ship::tick()
{
	// Work out the relative position of the pointer
	Vect3d pointerPos(manager->pointerx, manager->pointery, 0);
	Vect3d relPos = pointerPos - location;
	if (relPos.magnitudeSquared() > (MinPointerDist * MinPointerDist)) {
		// find angle of pointer
		Component targetAngle = relPos.angle();
		Component angleDiff = targetAngle - azimuth;
		normaliseAngle(angleDiff);
		if (std::fabs(angleDiff) > AngleTolerance) {
			if (angleDiff > 0) {
				azimuth += TurnSpeed;
			} else {
				azimuth -= TurnSpeed;
			}
		} else {
			// Within tolerance.
			azimuth = targetAngle;
		}
		normaliseAngle(azimuth);
	}
	
	// Sample mouse buttons
	engineOn = ((mouse_b & 1) != 0);
	tractorOn = ((mouse_b & 2) != 0);
	
	if (playerInfo.fuel <= 0) {
		engineOn = false;
	}
	
	if (engineOn) {
		// Accelerate appropriately.
		Vect3d accel = Vect3d::fromPolar(azimuth, EngineAccel);
		velocity = velocity + accel;
		createExhaustParticles();
		playerInfo.fuel --;
	}
	if (tractorOn) {
		applyTractorBeam();
	}
	
	location = location + velocity;
	ApplyFriction(velocity, 0.015);
	tractorOffset += TractorTextureSpeed;
	if (tractorOffset > 1.0) {
		tractorOffset -= 1.0;
	}
}

// Utility function
void ApplyFriction(Vect3d & velocity, double amount)
{
	if (velocity.magnitudeSquared() > (MinFrictionSpeed * MinFrictionSpeed)) {
		// Decrease by a proportion.
		velocity.scalarprod(1.0 - amount);
	} else {
		velocity = Vect3d(0,0,0);
	}
}
			
void Ship::collided(GameObject * what)
{
	// Create a big explosion.
	dead = true;
	manager->createExplosion(this);
	playerInfo.dead = true;
}

